import numpy as np

from . import mat_index_comb

from numpy.typing import NDArray

def p_dcca_matrix_to_simple(matrix_pdcca:NDArray[np.float64]) ->NDArray[np.float64]:
    """_summary_

    Args:
        matrix_pdcca (NDArray[np.float64]): _description_

    Returns:
        NDArray[np.float64]: _description_
    """    
    series_count = matrix_pdcca.shape[0]
    tws_count = matrix_pdcca.shape[2]
    DCCA_of = mat_index_comb(series_count)
    shape = (tws_count, DCCA_of.shape[0])
   

    P_DCCA_arr = np.full(shape=shape,fill_value=np.nan, dtype=matrix_pdcca.dtype)

    for n_index in range(tws_count):
        for i, pair in enumerate(DCCA_of):
            P_DCCA_arr[n_index, i] = matrix_pdcca[pair[0], pair[1], n_index] 

    return P_DCCA_arr



import numpy as np
from typing import Any
from numpy.typing import NDArray

def arr_2d_to_c(arr:NDArray[Any])-> NDArray[Any]:
    """Convert a 2D numpy array to a array of pointers that can be passed to a C ABI funtion.
    Args:
        arr (NDArray[Any]): 2D array of any type.
    Returns:
        NDArray[Any]: 1D array of pointers.
    """
        # Ensure the array is C-contiguous
    arr = np.ascontiguousarray(arr)
    
    # Safely access the base address
    base_address = arr.__array_interface__['data']
    if isinstance(base_address, tuple):  # Handle older NumPy versions
        base_address = base_address[0]
    
    # Compute row pointers
    row_pointers = (base_address + np.arange(arr.shape[0]) * arr.strides[0]).astype(np.uintp)
    
    return np.ascontiguousarray(row_pointers)


import numpy as np
from typing import Literal

from numpy.typing import NDArray

from . import detrended_series, mat_index_comb

from .p_dcca_simple_output import p_dcca_simple_output
from .p_dcca_matrix_output import p_dcca_matrix_output

ENUM_DCCA_of = Literal['all']

# P_DCCA calculator
def p_dcca(
    data: NDArray[np.float64], tws: NDArray[np.int64] | NDArray[np.float64], DCCA_of: np.ndarray | ENUM_DCCA_of ="all", P_DCCA_output_format="simple"
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
    # setting time_steps
    time_steps = np.arange(data.shape[0])
    if type(DCCA_of) == str:
        if DCCA_of == "all":
            DCCA_of = mat_index_comb(data, axis=1)
    # Global outputs for DFA and DCCA
    F_DFA_arr = np.zeros(shape=(tws.shape[0], data.shape[1]), dtype=data.dtype)
    DCCA_arr = np.zeros(shape=(tws.shape[0], DCCA_of.shape[0]), dtype=data.dtype)
    # P_DCCA global output format and function
    # simple output
    if P_DCCA_output_format == "simple":
        # output array
        P_DCCA_arr = np.full(shape=(tws.shape[0], DCCA_of.shape[0]) ,fill_value=np.nan, dtype=data.dtype)
        # output function
        P_DCCA_output_funtion = p_dcca_simple_output
    # matrix output
    elif P_DCCA_output_format == "matrix":
        # output matrix
        P_DCCA_arr = np.full(
            shape=(DCCA_of.max() + 1, DCCA_of.max() + 1, tws.shape[0]),fill_value=np.nan, dtype=data.dtype
        )
        # fill diagonal with ones
        r = np.arange(DCCA_of.max() + 1)
        P_DCCA_arr[r,r, :] = 1
        del r
        # output function
        P_DCCA_output_function = p_dcca_matrix_output
    # for time scales in n
    for n_index, n  in enumerate(tws):
        # in time scale (n) accumulators
        f2dfa_n = np.full(shape=(data.shape[0] - n, data.shape[1]),fill_value=np.nan, dtype=data.dtype)
        dcca_n = np.full(shape=(data.shape[0] - n, DCCA_of.shape[0]), fill_value=np.nan, dtype=data.dtype)
        detrended_mat = np.full(shape=(n + 1, data.shape[1]), fill_value=np.nan, dtype=data.dtype)

        for i in range(data.shape[0] - n): # for each box

            detrended_series( # inputs
                time_steps[i : i + (n + 1)],  # arr_x
                data[i : i + (n + 1), :],  # mat_y
                detrended_mat,  # output
            )

            f2dfa_n[i] = np.power(detrended_mat, 2).mean(axis=0)

            for j, pair in enumerate(DCCA_of):

                dcca_n[i, j] = (
                    detrended_mat[:, pair[0]] * detrended_mat[:, pair[1]]
                ).mean(axis=0)

        F_DFA_arr[n_index, :] = np.sqrt(f2dfa_n.mean(axis=0))

        DCCA_arr[n_index, :] = dcca_n.mean(axis=0)

        # calculation of P_DCCA
        P_DCCA_output_function(n_index, DCCA_of, F_DFA_arr, DCCA_arr, P_DCCA_arr)

    return F_DFA_arr, DCCA_arr, P_DCCA_arr



import platform
import ctypes
from pathlib import Path

import numpy as np
from typing import Literal

from . import mat_index_comb

from .array_to_c_pointer_convert import arr_2d_to_c

from numpy.typing import NDArray
from numpy.ctypeslib import ndpointer

uint_c_type = np.uintp

ENUM_DCCA_of = Literal['all']

def p_dcca(input_data: NDArray[np.float64], 
           tws:  NDArray[np.int64] | NDArray[np.float64], 
           DCCA_of: np.ndarray | ENUM_DCCA_of ="all", 
           P_DCCA_output_matrix: bool=False
        ) ->    tuple[
                NDArray[np.float64],    # DFA
                NDArray[np.float64],    # DCCA
                NDArray[np.float64]     # P_DCCA
                ]:

    """A function that calculates the 
       <span>&Rho;<sub>DCCA</sub></span>
        for a group of time series

        Args:
            input_data (NDArray[np.float64]): 2D array of times series integrated data.
            tws (NDArray[np.float64] | NDArray[np.float64]): 1D array of time scales.
            DCCA_of (np.ndarray | list | None, optional): _description_. Defaults to None.
            P_DCCA_output_matrix (bool, optional): _description_. Defaults to False.


        Returns:
            <span>A tuple of 3 matrices:</span>
            DFA(NDArray[np.float64]):_description_. 
            DCCA(NDArray[np.float64]):_description_.
            <span>&Rho;<sub>DCCA</sub></span>(NDArray[np.float64]):_description_.
    """
    
    assert (tws[:-1] < tws[1:]).all() == True , ("""time window scales (tws) values must be in crescent order.""")


    # setting lib path

    sys_dict = {'Windows': 'zebendezig.dll', 'Linux': 'libzebendezig.so', 'Darwin': 'libzebendezig.dylib'}

    sys_info = platform.uname()

    lib_folder = Path(__file__).parent / 'zig_libs'

    lib_file = lib_folder / sys_dict[sys_info.system]

    zz = ctypes.cdll.LoadLibrary(lib_file)

    # outputs types
    c_2d_any_1d_uint = ndpointer(dtype = uint_c_type, ndim=1, flags='C')
    c_1d_double = ndpointer(dtype=np.double, ndim=1, flags='C')
    # getting the module file
    _p_dcca = zz.p_dcca
    # setting outputs
    _p_dcca.argtypes = [ 
                    c_1d_double, ctypes.c_size_t, ctypes.c_size_t, # input data
                    c_2d_any_1d_uint, ctypes.c_size_t, # tws
                    c_1d_double, # Time steps
                    c_2d_any_1d_uint, ctypes.c_size_t, # DCCA of
                    # outputs
                    c_2d_any_1d_uint, # DFA_arr
                    c_2d_any_1d_uint, # DCCA_arr
                    c_1d_double, ctypes.c_bool, # DCCA_arr
                    ]

    # Preparing input data    

    if not tws.flags.c_contiguous:
        tws = np.ascontiguousarray(tws)
    
    # preparing DCCA_ot array

    if type(DCCA_of) == str:
        if DCCA_of == "all":
            DCCA_of =  np.ascontiguousarray(mat_index_comb(input_data, axis=1))
    # ensuring data compatibility
    DCCA_of = DCCA_of.astype(uint_c_type)
    c_DCCA_of = arr_2d_to_c(DCCA_of)

    
    # preparing tws array
    tws = tws.astype(uint_c_type)

    # preparing output array
    F_DFA_arr = np.ascontiguousarray(np.zeros(shape=(tws.shape[0], input_data.shape[1]), dtype=input_data.dtype))
    c_DFA_arr = arr_2d_to_c(F_DFA_arr)
    DCCA_arr = np.ascontiguousarray(np.zeros(shape=(tws.shape[0], DCCA_of.shape[0]), dtype=input_data.dtype))
    c_DCCA_arr = arr_2d_to_c(DCCA_arr)

    # preparing data
    data_shape = input_data.shape
    try:
        x_len = data_shape[0]
        x_cnt = data_shape[1]
    except:
        x_len = input_data.size
        x_cnt = 1
    input_data = np.ascontiguousarray(np.asfortranarray(input_data).flatten(order="K"))

    time_steps = np.ascontiguousarray(np.arange(x_len, dtype=input_data.dtype))

    # preparing P_DCCA array
    # output array
    if P_DCCA_output_matrix == False:
        P_DCCA_shape = np.array([tws.shape[0], DCCA_of.shape[0]])
        P_DCCA_arr = np.ascontiguousarray(np.full(shape=(P_DCCA_shape.prod()) ,fill_value=np.nan, dtype=input_data.dtype))
    # output matrix
    elif P_DCCA_output_matrix == True:
        P_DCCA_arr = np.full(
            shape=(x_cnt, x_cnt, tws.shape[0]),fill_value=np.nan, dtype=input_data.dtype)
        # fill diagonal with ones
        P_DCCA_shape =  P_DCCA_arr.shape
        r = np.arange(x_cnt, dtype= np.int64)
        P_DCCA_arr[r,r, :] = 1
        del r
        # flatteing matrix
        P_DCCA_arr = np.ascontiguousarray(P_DCCA_arr.flatten())

    # calling functon  
    _p_dcca(input_data,  x_len, x_cnt, 
            tws, tws.size, 
            time_steps, 
            c_DCCA_of , DCCA_of.shape[0],
            # Outputs
            c_DFA_arr,
            c_DCCA_arr,
            P_DCCA_arr,
            # P_DCCA output matrix
            P_DCCA_output_matrix
            )

    # reshaping P_DCCA output
    P_DCCA_arr = P_DCCA_arr.reshape(P_DCCA_shape)

    return [F_DFA_arr, DCCA_arr, P_DCCA_arr]



from typing import (Literal)

import numpy as np
from numpy.typing import (NDArray)


def dmcx2_from_p_dcca_matrix(P_DCCA_arr: NDArray[np.float64],  dmcx2_of: NDArray[np.float64]) -> NDArray[np.float64]:
    """_summary_

    Args:
        P_DCCA_arr (NDArray[np.float64]): _description_
        dmcx2_of (NDArray[np.float64]): _description_

    Returns:
        NDArray[np.float64]: _description_
    """    
    # DMCx2 output matrix
    DMCx2_arr = np.full(shape=(P_DCCA_arr.shape[2], dmcx2_of.shape[0]), fill_value=np.nan, dtype=P_DCCA_arr.dtype)

    for n_index in range(P_DCCA_arr.shape[2]):
        P_DCCA_arr_2D = P_DCCA_arr[:, :, n_index]
        for j,  dmcx2_of_1D in enumerate(dmcx2_of):
            DMCx2_arr[n_index, j] = dmcx2_from_p_dcca_matrix_2d(P_DCCA_arr_2D,  dmcx2_of_1D)
    return DMCx2_arr


def dmcx2_from_p_dcca_matrix_2d(P_DCCA_arr_2D: NDArray[np.float64],  dmcx2_of_1D: NDArray[np.float64]) -> NDArray[np.float64]:
    """Calculates the 

    Args:
        P_DCCA_arr_2D (NDArray[np.float64]): 2D array of <span>&Rho;<sub>DCCA</sub></span> results 
        dmcx2_of__1D (NDArray[np.float64]): 1D array of indexes

    Returns:
        <span>DMC<sub>x</sub><sup>2</sup></span>([float64]): _description_
    """    
    y_index = dmcx2_of_1D[0:1]
    x_indexes = dmcx2_of_1D[1:]

    mat_x = P_DCCA_arr_2D[np.ix_(x_indexes, x_indexes)]
    vec_y = P_DCCA_arr_2D[np.ix_(x_indexes, y_index)]

    return vec_y.T @ np.linalg.inv(mat_x) @ vec_y


import numpy as np
from numpy.typing import NDArray


def ordering_x_dmcx2_of(dmcx2_of:NDArray[np.int64])-> NDArray[np.int64]:
    """Ordering the independent indexes of the dmcx2_of NDArray

    Args:
        dmcx2_of (NDArray[np.int64]): dmcx2_of NDArray

    Returns:
        NDArray[np.int64]: dmcx2_of NDArray with the independent variables ordered.
    """
    print(type(dmcx2_of))
    if   type(dmcx2_of) == np.ndarray:
        if dmcx2_of.ndim == 2:
            y_serie = dmcx2_of[:, 0]
            x_series = np.sort(dmcx2_of[:, 1:], axis=1)
            out = np.c_[y_serie, x_series]

        elif dmcx2_of.ndim == 1:
            y_serie = dmcx2_of[0:1]
            x_series = np.sort(dmcx2_of[1:])
            out =  np.concatenate((y_serie, x_series))
        else:
            print("dmcx2_of must be a 2D or 1D Numpy array")

    return out



import numpy as np

def p_dcca_simple_output(n_index, DCCA_of, F_DFA_arr, DCCA_arr, P_DCCA_arr):

    for i, dcca_pair in enumerate(DCCA_of):

                P_DCCA_arr[n_index, i] = DCCA_arr[n_index, i] / (F_DFA_arr[n_index, dcca_pair[0]] * F_DFA_arr[n_index, dcca_pair[1]])
    


from typing import (Literal, Any, Union)

import numpy as np
from numpy.typing import (NDArray, DTypeLike)


from . import (
    dcca_of_from_dmcx2_of,
    dmc_of_all_as_y,
    dmcx2_from_p_dcca_matrix,
    p_dcca_pure_python,
)




def dmcx2_pure_python(input_data: NDArray[np.float64], 
          tws: NDArray[np.int64] | NDArray[np.float64], 
          dmcx2_of: NDArray[np.float64] | Literal['all-full', 'first-full']  = 'all-full'
           )-> tuple[
                                                            NDArray[np.float64],
                                                            NDArray[np.float64],
                                                            NDArray[np.float64],
                                                            NDArray[np.float64]
                                                             ]:

    """
        A function that calculates the <span>DMC<sub>x</sub><sup>2</sup></span> for a group of time series

        Args:
            input_data (NDArray[np.float64]): _description_.
            tws (NDArray[np.int64] | NDArray[np.float64]): _description_.
            dmcx2_of (NDArray[np.float64] | Literal['all-full', 'first-full'], optional): _description_. Defaults to 'all-full'.
            DCCA_of (np.ndarray | None, optional): _description_. Defaults to None.

        Returns:
            <span>A tuple of 4 matrices:</span><br>
            DFA(NDArray[np.float64]):_description_,<br>
            DCCA(NDArray[np.float64]):_description_,<br>
            <span>&Rho;<sub>DCCA</sub></span>(NDArray[np.float64]):_description_,<br>
            <span>DMC<sub>x</sub><sup>2</sup></span>(NDArray[np.float64]):_description_.<br>

    """
 
    
    if type(dmcx2_of) == str:

        # creating ndarray of y and x values for DMCx2 calculations
        if dmcx2_of == 'first-full':
            dmcx2_of = np.array( [np.arange(input_data.shape[1])])
        # creating ndarray of y and x values for DMCx2 calculations
        elif dmcx2_of == 'all-full':
            dmcx2_of = dmc_of_all_as_y(input_data)
    
    test_dmcx2_of = dmcx2_of[:,1:]
    assert (test_dmcx2_of[:,:-1] < test_dmcx2_of[:,1:]).all() == True , ("""
Dmcx2 x values out of order: use zebende.ordering_x_dmcx2_of(dmcx2_of) to fix it before passing the dmcx2_of value to zebende.dmcx2() function""")
    del test_dmcx2_of

    # creating ndarray for P_DCCA calculations based on the DMCx2 array
    DCCA_of = dcca_of_from_dmcx2_of(dmcx2_of)


    # P_DCCA calculations
    F_DFA_arr, DCCA_arr, P_DCCA_arr = p_dcca_pure_python(input_data=input_data, tws=tws, DCCA_of=DCCA_of,  P_DCCA_output_matrix = True)

    # DMCx2 output matrix
    DMCx2_arr = np.full(shape=(tws.shape[0], dmcx2_of.shape[0]), fill_value=np.nan, dtype=input_data.dtype)

    DMCx2_arr = dmcx2_from_p_dcca_matrix(P_DCCA_arr, dmcx2_of)

    return F_DFA_arr, DCCA_arr, P_DCCA_arr, DMCx2_arr



import numpy as np

def p_dcca_matrix_output(n_index:int, DCCA_of, F_DFA_arr, DCCA_arr, P_DCCA_arr):  

    for i in range(DCCA_of.shape[0]):
            p_dcca = DCCA_arr[n_index, i] / (F_DFA_arr[n_index, DCCA_of[i][0]] * F_DFA_arr[n_index, DCCA_of[i][1]])
            P_DCCA_arr[DCCA_of[i, 0], DCCA_of[i, 1], n_index] = p_dcca
            P_DCCA_arr[DCCA_of[i, 1], DCCA_of[i, 0], n_index] = p_dcca



from itertools import combinations
from numpy.typing import NDArray

import numpy as np


def dcca_of_from_dmcx2_of(dmc_of:NDArray[np.int64]) -> NDArray[np.int64]:
    """_summary_

    Args:
        dmc_of (NDArray[np.int64]): _description_

    Returns:
        NDArray[np.int64]: _description_
    """    
    if type(dmc_of) == np.ndarray:
        dmc_of = dmc_of.tolist()
    out = []
    for i in range(len(dmc_of)):
        temp = list(combinations(dmc_of[i], 2))
        for j in temp:
            j = list(j)

            j.sort()

            if j not in out:
                out.append(j)
    return np.array(out)



import numpy as np
from numpy.typing import NDArray

# integrates series
def integrated_series(mat_series:NDArray[np.float64]) -> NDArray[np.float64]:
    """Returns a matrix of integrates series from a matrix of time series.
        The integrates series is a cumulative sum of the values of the series subtracted by the mean.

    Args:
        mat_series (NDArray[np.float64]): Matrix of time series with one serie per column.

    Returns:
        NDArray[np.float64]: Matrix of integrated time series with one integrated time series per column.
    """    
    out = (mat_series - mat_series.mean(axis=0)).cumsum(axis=0)
    return out



import numpy as np
from . import ordering_x_dmcx2_of
from numpy.typing import NDArray


def dmc_of_all_as_y(data:NDArray[np.float64]) -> NDArray[np.float64]:
    """Generates a 2D array for  <span>DMC<sub>x</sub><sup>2</sup></span> calculations

    Args:
        data (NDArray[np.float64]): 2D array of data.

    Returns:
        NDArray[np.float64]: 2D array where in each row, the first element is the dependent variable,
        and the other valus are the intependent ones.
        In each line one of the series is choosen as the dependent variable,
        so the number of lines correspond to the number of columns in the data array.
    """    
    aux = list(range(data.shape[1]))
    dmc_list = [aux]
    for i in range(data.shape[1]-1):
        aux = aux[1:] + aux[:1]


        dmc_list.append(aux)

    dmc_list = np.array(dmc_list)

    dmc_list = ordering_x_dmcx2_of(dmc_list)

    return dmc_list




import numpy as np
from . import mat_index_comb
from numpy.typing import NDArray

def p_dcca_simple_to_matrix(simple_pdcca:NDArray[np.float64], DCCA_of:NDArray[np.int64] | None = None ) ->NDArray[np.float64]:

    """Convert a simple (table) of  <span>&Rho;<sub>DCCA</sub></span> values to the matrix output.

    Args:
        simple_pdcca:NDArray[np.float64]: Table of <span>&Rho;<sub>DCCA</sub></span> outputs.
        DCCA_of (np.ndarray | None, optional): The DCCA_of used for the <span>&Rho;<sub>DCCA</sub></span> calculations.
        If None, the function will assume it was created with the Default option in the p_dcca function. Defaults to None.

    Returns:
        NDArray[np.float64]: 3D matrix with the values of the <span>&Rho;<sub>DCCA</sub></span> of each time scale as a level.
    """
    
    if DCCA_of is None:
        series_count = int(np.round((1 + np.sqrt(1 + 8 *simple_pdcca.shape[1]))/2, 0))
    
        DCCA_of = mat_index_comb(series_count)
    else:
        series_count = DCCA_of.flatten().max()+1

    tws_count = simple_pdcca.shape[0]

    shape = (series_count, series_count, tws_count)
    P_DCCA_arr = np.full(shape=shape,fill_value=np.nan, dtype=simple_pdcca.dtype)
    # fill diagonal with ones
    r = np.arange(series_count, dtype= np.int64)
    P_DCCA_arr[r,r, :] = 1
    del r

    for n_index in range(tws_count):
        for i, pair in enumerate(DCCA_of):
            P_DCCA_arr[pair[0], pair[1], n_index] = simple_pdcca[n_index, i]
            P_DCCA_arr[pair[1], pair[0], n_index] = simple_pdcca[n_index, i]

    return P_DCCA_arr



from itertools import combinations

import numpy as np

from numpy.typing import NDArray

def mat_index_comb(mat:NDArray[np.float64] | int, axis:int = 1) -> NDArray[np.uintp] | None:

    """_summary_

    Args:
        mat (NDArray[np.float64], int ): Matrix of time series with one serie per column.
        axis (int , optional):

    Returns:
        (NDArray[np.uintp], None): _description_
    """
    if type(mat) == np.ndarray:
        return np.array(list(combinations(range(mat.shape[axis]), 2)), dtype=np.uintp)
    elif type(mat) == int:
        return np.array(list(combinations(range(mat), 2)), dtype=np.uintp)
    else:
        print('mat of type {} not supported'.format(type(mat)))
        return None



from .detrended_series import liner_ls_fit as liner_ls_fit
from .detrended_series import detrended_series as detrended_series
from .detrended_series import detrended_fit_series as detrended_fit_series

from .integrated_series import integrated_series as integrated_series
from .time_windows import time_windows as time_windows

from .mat_index_comb import mat_index_comb as mat_index_comb

from .p_dcca import p_dcca as p_dcca

from .dfa import dfa as dfa
from .dcca_of_from_dmcx2_of import dcca_of_from_dmcx2_of as dcca_of_from_dmcx2_of

from .p_dcca_simple_output import p_dcca_simple_output as p_dcca_simple_output
from .p_dcca_matrix_output import p_dcca_matrix_output as p_dcca_matrix_output

from .p_dcca_simple_to_matrix import p_dcca_simple_to_matrix as p_dcca_simple_to_matrix
from .p_dcca_matrix_to_simple import p_dcca_matrix_to_simple as p_dcca_matrix_to_simple

from .ordering_x_dmcx2_of import ordering_x_dmcx2_of as ordering_x_dmcx2_of

from .dmc_of_all_as_y import dmc_of_all_as_y as dmc_of_all_as_y
from .dmcx2_from_p_dcca_matrix import dmcx2_from_p_dcca_matrix as dmcx2_from_p_dcca_matrix
from .dmcx2_from_p_dcca_matrix import dmcx2_from_p_dcca_matrix_2d as dmcx2_from_p_dcca_matrix_2d
from .dmcx2_pure_python import dmcx2_pure_python as dmcx2_pure_python
from .dmcx2 import dmcx2 as dmcx2

from .print_pdcca_matrix_by_level import print_pdcca_matrix_by_level as print_pdcca_matrix_by_level





import numpy as np
from numpy.typing import NDArray


def print_pdcca_matrix_by_level(pdcca:NDArray[np.float64]):
    """Print the <span>&Rho;<sub>DCCA</sub></span> matrix by level.

    Args:
        pdcca (NDArray[np.float64]): <span>&Rho;<sub>DCCA</sub></span> calculation in the matrix format.
    """

    print(pdcca.transpose(2,0,1))


import numpy as np

def detrended_fit_series(arr_x, arr_y, out):
    
    n = arr_x.size
    col = arr_y.shape[1]
    for j in range(col):
        x_sum = 0
        y_sum = 0
        xy_sum = 0
        x2_sum = 0
        for i in range(n):
            x_sum += arr_x[i]
            y_sum += arr_y[i,j]
            xy_sum += arr_x[i] * arr_y[i,j]
            x2_sum += arr_x[i]**2
        # slope (m in y = mx +b)
        slope = (
                ((n * xy_sum) - (x_sum * y_sum)) /
                ((n * x2_sum) - (x_sum**2))
                )
        # intercept (b in y = mx +b)
        inter = (
                (y_sum - (slope * x_sum)) /
                (n)
                )
        
        for i in range(n):
            out[i,j] = arr_y[i,j] - (slope * arr_x[i] + inter)
    return out



# linaer Least Squares fit
def liner_ls_fit(arr_x, arr_y):
    n = arr_x.size

    x_sum = 0
    y_sum = 0
    xy_sum = 0
    x2_sum = 0
    for i in range(n):
        x_sum += arr_x[i]
        y_sum += arr_y[i,:]
        xy_sum += arr_x[i] * arr_y[i,:]
        x2_sum += arr_x[i]**2
    # slope (m in y = mx +b)
    slope = (
            ((n * xy_sum) - (x_sum * y_sum)) /
            ((n * x2_sum) - (x_sum**2))
            )
    # intercept (b in y = mx +b)
    inter = (
            (y_sum - (slope * x_sum)) /
            (n)
            )

    return slope, inter


# detrended series
def detrended_series(arr_x, arr_y, out):
    n = arr_x.size
    slope, inter = liner_ls_fit(arr_x, arr_y)

    for i in range(n):
        out[i] = arr_y[i,:] - (slope * arr_x[i] + inter)




import numpy as np

from numpy.typing import NDArray

from . import detrended_series, mat_index_comb, detrended_fit_series

from .p_dcca_simple_output import p_dcca_simple_output
from .p_dcca_matrix_output import p_dcca_matrix_output


# P_DCCA calculator
def dfa(data: NDArray[np.float64], tws: NDArray[np.int64]) -> NDArray[np.float64]:
    
    """Calculates the Detrended Fluctuation analysis (DFA) for one or more time series. 

    Args:
        data (NDArray[np.float64]): data matrix with one series per column.
        tws (NDArray[np.int64]): 1D array of time windows.

    Returns:
        NDArray[np.float64]: DFA table.
    """ 
        # setting time_steps in None is passed

    time_steps = np.arange(data.shape[0])


    # Global outputs
    F_DFA_arr = np.zeros(shape=(tws.shape[0], data.shape[1]), dtype=data.dtype)


    # for time scales in n
    for n_index in range(len(tws)):

        n = tws[n_index]

        # in time scale (n) accumulators


        f2dfa_n = np.full(shape=(data.shape[0] - n, data.shape[1]),fill_value=np.nan, dtype=data.dtype)

        detrended_mat = np.full(shape=(n + 1, data.shape[1]), fill_value=np.nan, dtype=data.dtype)
    
        # Operações dentro das caixas (sobrepostas)

        for i in range(data.shape[0] - n):

            # 2-- dividir o sinal em N-n janelas temporais de comprimento n
            # janela temporal

            # 3-- Ajustar uma curva de tedência

            # geralente polinômio de primerio grau

            detrended_series( # inputs
                time_steps[i : i + (n + 1)],  # arr_x
                data[i : i + (n + 1), :],  # mat_y
                detrended_mat,  # output
            )

            f2dfa_n[i] = np.power(detrended_mat, 2).mean(axis=0)



        # 5--para cada escala temporal
        # salvar valor de cada escala temporal

        F_DFA_arr[n_index, :] = np.sqrt(f2dfa_n.mean(axis=0))



    return F_DFA_arr



import numpy as np
from numpy.typing import NDArray


# time window array
def time_windows(n_min: int, n_max: int, exp_fac:float = 8.0, ensure_max:bool = True) -> NDArray[np.int64]:

    """Generate an ordered array of time sclaes for the Detrended calculations.
    Args:
        n_min (int): Minimun value for a time window (box).
        n_max (int): Maximun value for a time window (box).
        exp_fac (float, optional): Exponencial factor to calculate successive time windows in the array. Defaults to 8.0.
        ensure_max (bool, optional): Ensure that the Maximun value is. Defaults to True.
    Returns:
        NDArray[np.int64]: Time scales array.
    """
    n = n_min
    tmp = []
    ir = 0
    while n <= n_max:
        tmp.append(n)
        ir = ir + 1
        n = int((n_min + ir) * np.power(np.power(2, 1.0 / exp_fac), ir))
    if ((ensure_max == True) and (tmp[-1] < n_max)):
        tmp.append(n_max)

    return np.array(tmp, dtype=np.int64)
